#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>

#include "bonnie.h"
#include "bon_file.h"

CPCCHAR rand_chars = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";


COpenTest::COpenTest()
{
  m_number = 0;
  m_used = NULL;
  m_dirname = NULL;
  m_file_name_buf = NULL;
  m_file_names = NULL;
  m_mask_bits = 1;
  unsigned int tmp = 1;
  while( (tmp = tmp << 1) != (8 * sizeof(MASK_TYPE)) )
    m_mask_bits++;
  m_last_mask = (1 << m_mask_bits) - 1;
}

int COpenTest::get_random(bool set)
{
  MASK_TYPE bitcheck = set ? 0 : 1;
  MASK_TYPE wordcheck = set ? 0 : ~0;
  int rand = random() % m_number;
  int randShift = rand & m_last_mask;
  rand = rand >> m_mask_bits;
  for( ; m_used[rand] == wordcheck; rand++)
  {
    if(rand == m_used_size - 1)
      rand = -1;
  }
  MASK_TYPE w = (m_used[rand] >> randShift)
               | (m_used[rand] << ((8 * sizeof(MASK_TYPE)) - randShift));
  while((w & 1) == bitcheck)
  {
    w = w >> 1;
    randShift++;
  }
  if(randShift >= int(8 * sizeof(MASK_TYPE)) )
    randShift -= (8 * sizeof(MASK_TYPE));
  return rand << m_mask_bits | randShift;
}

COpenTest::~COpenTest()
{
  if(m_used)
  {
    fprintf(stderr, "Cleaning up test directory after error.\n");
    for(int i = 0; i < m_number; i++)
    {
      if(test_used(i))
        unlink(m_file_names[i]);
    }
    chdir("..");
    delete m_used;
    m_used = NULL;
  }
  if(m_dirname)
  {
    rmdir(m_dirname);
    delete m_dirname;
  }
  delete m_file_name_buf;
  delete m_file_names;
}

void COpenTest::make_names(bool do_random)
{
  delete m_file_name_buf;
  delete m_file_names;
  m_file_name_buf = new char[(MaxNameLen + 1) * m_number];
  m_file_names = new PCHAR[m_number];
  PCHAR buf = m_file_name_buf;
  for(int i = 0; i < m_number; i++)
  {
    char rand_buf[RandExtraLen + 1];
    int len = random() % (RandExtraLen + 1);
    int j;
    for(j = 0; j < len; j++)
    {
      rand_buf[j] = rand_chars[random() % strlen(rand_chars)];
    }
    rand_buf[j] = '\0';
    if(do_random)
    {
      sprintf(buf, "%s%07d", rand_buf, i);
    }
    else
    {
      sprintf(buf, "%07d%s", i, rand_buf);
    }
    m_file_names[i] = buf;
    buf += strlen(buf) + 1;
  }
}

int COpenTest::create(CPCCHAR dirname, BonTimer &timer, int num, int max_size
                    , int min_size, bool do_random)
{
  m_number = num * DirectoryUnit;
  make_names(do_random);
  m_max = max_size;
  m_min = min_size;
  m_size_range = m_max - m_min;
  m_dirname = new char[strlen(dirname) + 1];
  strcpy(m_dirname, dirname);

  if(mkdir(dirname, S_IRWXU))
  {
    fprintf(stderr, "Can't make directory %s\n", dirname);
    return -1;
  }
  if(chdir(dirname))
  {
    fprintf(stderr, "Can't change to directory %s\n", dirname);
    return -1;
  }
  m_used_size = m_number >> m_mask_bits;
  m_used = new MASK_TYPE[m_used_size];
  char buf[Chunk];
  int i;
  for(i = 0; i < (m_number >> m_mask_bits); i++)
    m_used[i] = 0;

  timer.timestamp();
  for(i = 0; i < m_number; i++)
  {
    int fd = creat(m_file_names[i], S_IRUSR | S_IWUSR);
    if(fd == -1)
    {
      fprintf(stderr, "Can't create file %s\n", m_file_names[i]);
      return -1;
    }
    if(m_max)
    {
      int size;
      if(m_size_range)
        size = m_min + (random() % (m_size_range + 1));
      else
        size = m_max;
      for(int i = 0; i < size; i += Chunk)
      {
        int to_write = size - i;
        if(to_write > Chunk) to_write = Chunk;
        if(to_write != write(fd, static_cast<void *>(buf), to_write))
        {
          fprintf(stderr, "Can't write data.\n");
          return -1;
        }
      }
    }
    close(fd);
    set_used(i);
  }
  timer.get_delta_t(do_random ? CreateRand : CreateSeq);
  return 0;
}

int COpenTest::delete_random(BonTimer &timer)
{
  timer.timestamp();
  int i;
  for(i = 0; i < m_number; i++)
  {
    int j = get_random(true);
    if(unlink(m_file_names[j]))
    {
      fprintf(stderr, "Can't delete file %s\n", m_file_names[j]);
      return -1;
    }
    set_unused(j);
  }
  delete m_used;
  m_used = NULL;
  chdir("..");
  rmdir(m_dirname);
  delete m_dirname;
  m_dirname = NULL;
  timer.get_delta_t(DelRand);
  return 0;
}

int COpenTest::delete_sequential(BonTimer &timer)
{
  timer.timestamp();
  DIR *d = opendir(".");
  if(!d)
  {
    fprintf(stderr, "Can't open directory.\n");
    return -1;
  }
  dirent *file_ent;

  int count = 0;
  while((file_ent = readdir(d)) != NULL)
  {
    if(file_ent->d_name[0] != '.')
    {
      if(unlink(file_ent->d_name))
      {
        fprintf(stderr, "Can't delete file %s\n", file_ent->d_name);
        return -1;
      }
      count++;
    }
  }
  delete m_used;
  m_used = NULL;
  chdir("..");
  rmdir(m_dirname);
  delete m_dirname;
  m_dirname = NULL;
  closedir(d);
  if(count != m_number)
  {
    fprintf(stderr, "Expected %d files but only got %d\n", m_number, count);
    return -1;
  }
  timer.get_delta_t(DelSeq);
  return 0;
}

int COpenTest::stat_file(CPCCHAR file)
{
  struct stat st;
  if(stat(file, &st))
  {
    fprintf(stderr, "Can't stat file %s\n", file);
    return -1;
  }
  if(st.st_size)
  {
    int fd = open(file, O_RDONLY);
    if(fd == -1)
    {
      fprintf(stderr, "Can't open file %s\n", file);
      return -1;
    }
    char buf[Chunk];
    for(int i = 0; i < st.st_size; i += Chunk)
    {
      int to_read = st.st_size - i;
      if(to_read > Chunk) to_read = Chunk;
      if(to_read != read(fd, static_cast<void *>(buf), to_read))
      {
        fprintf(stderr, "Can't read data.\n");
        return -1;
      }
    }
    close(fd);
  }
  return 0;
}

int COpenTest::stat_random(BonTimer &timer)
{
  timer.timestamp();

  int i;
  for(i = 0; i < m_number; i++)
  {
    if(-1 == stat_file(m_file_names[random() % m_number]))
      return -1;
  }
  timer.get_delta_t(StatRand);
  return 0;
}

int COpenTest::stat_sequential(BonTimer &timer)
{
  timer.timestamp();
  DIR *d = opendir(".");
  if(!d)
  {
    fprintf(stderr, "Can't open directory.\n");
    return -1;
  }
  dirent *file_ent;
  int count = 0;
  while((file_ent = readdir(d)) != NULL)
  {
    if(file_ent->d_name[0] != '.') // our files do not start with a dot
    {
      if(-1 == stat_file(file_ent->d_name))
        return -1;
      count++;
    }
  }
  closedir(d);
  if(count != m_number)
  {
    fprintf(stderr, "Expected %d files but only got %d\n", m_number, count);
    return -1;
  }
  timer.get_delta_t(StatSeq);
  return 0;
}

